import os
import sys
import multiprocessing
from multiprocessing.managers import ValueProxy
import time


def run(file_path: str, refresh_interval: float = 0.1):
    """
    run a python file and reload it whenever it is modified
    """

    file_closed = multiprocessing.Manager().Value("closed", False)

    if not os.path.isfile(file_path):
        base_path = os.path.dirname(sys.argv[0])
        if os.path.isfile(base_path + file_path):
            file_path = base_path + file_path
        elif os.path.isfile(base_path + os.sep + file_path):
            file_path = base_path + os.sep + file_path
        else:
            raise FileNotFoundError(
                "File not found: \n  "
                + file_path
                + "\n| "
                + base_path
                + file_path
                + "\n| "
                + base_path
                + os.sep
                + file_path
            )

    with open(file_path, "r") as f:
        code = f.read()

    p = multiprocessing.Process(target=_run, args=(code, file_closed))
    p.start()

    while not file_closed.value:
        with open(file_path, "r") as f:
            current_code = f.read()

        if code != current_code:
            print("Reloading code...")
            code = current_code

            p.terminate()
            p = multiprocessing.Process(target=_run, args=(code, file_closed))
            p.start()

        time.sleep(refresh_interval)


def _run(code, close: "ValueProxy[bool]"):
    new_globals = globals()
    new_globals["__name__"] = "__main__"
    exec(code, new_globals)

    close.value = True


if __name__ == "__main__":
    print("Name is main")
    run(sys.argv[1], *({"refresh_interval": sys.argv[2]}) if len(sys.argv) > 2 else {})



from typing import Iterable


def assert_true(condition, error_message_appendix=None):
    if condition:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"condition is False{error_message_appendix}")

def assert_false(condition, error_message_appendix=None):
    if not condition:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"condition is True{error_message_appendix}")

def assert_equal(a, b, error_message_appendix=None):
    if a == b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} != {b}{error_message_appendix}")

def assert_not_equal(a, b, error_message_appendix=None):
    if a != b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} == {b}{error_message_appendix}")

def assert_is(a, b, error_message_appendix=None):
    if a is b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} is not {b}{error_message_appendix}")

def assert_greater(a, b, error_message_appendix=None):
    if a > b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} <= {b}{error_message_appendix}")

def assert_greater_equal(a, b, error_message_appendix=None):
    if a >= b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} < {b}{error_message_appendix}")

def assert_less(a, b, error_message_appendix=None):
    if a < b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} >= {b}{error_message_appendix}")

def assert_less_equal(a, b, error_message_appendix=None):
    if a <= b:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} > {b}{error_message_appendix}")

def assert_close(a, b, epsilon=1e-6, error_message_appendix=None):
    if abs(a - b) < epsilon:
        return True
    else:
        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"{a} != {b}{error_message_appendix}")

def assert_type(obj, type, error_message_appendix=None):
    if not isinstance(obj, type):
        if isinstance(type, Iterable):
            if len(type) > 0:
                type_str = type[-1].__name__
            if len(type) > 1:
                type_str = type[-2].__name__ + " or " + type_str
            for i in range(3, len(type) + 1):
                type_str = type[-i].__name__ + ", " + type_str
        else:
            type_str = type.__name__

        if error_message_appendix is None:
            error_message_appendix = ""
        else:
            error_message_appendix = f": {error_message_appendix}"
        raise AssertionError(f"object is not an instance of {type_str}{error_message_appendix}")
    return True

def assert_types(objects, types, error_message_appendix=None):
    for index, obj, type in enumerate(zip(objects, types)):
        try:
            assert_type(obj, type)
        except AssertionError as e:
            if isinstance(type, Iterable):
                if len(type) > 0:
                    type_str = type[-1].__name__
                if len(type) > 1:
                    type_str = type[-2].__name__ + " or " + type_str
                for i in range(3, len(type) + 1):
                    type_str = type[-i].__name__ + ", " + type_str
            else:
                type_str = type.__name__
                
            if error_message_appendix is None:
                error_message_appendix = ""
            else:
                error_message_appendix = f": {error_message_appendix}"
            raise AssertionError(
                f"object number {index + 1} is not an instance of {type_str}{error_message_appendix}"
            ) from e
    return True



from ._get_instances import get_instances
from ._get_names import get_names
from ._get_all_subclasses import get_all_subclasses
from ._get_types import get_types
from ._get_first_shared_inheritance import get_first_shared_inheritance


from math import exp, log


class PEMDAS:
    """
    Enum for the order of operations.
    """
    AS = 0 # Addition, Subtraction
    MD = 1 # Multiplication, Division (and Modulo)
    E = 2 # Exponentiation
    P = 3 # Parentheses


class Operator:
    order = PEMDAS.P
    print_pattern = "$a [operator] $b"
    """
    $a, $b, $c, ... will be replaced by their values based on the _eval method's signature.
    """

    # TODO: Add proper min and max values
    def __init__(self):
        self.min_values = {}
        self.max_values = {}
    
    def __new__(cls, *args, **kwargs):
        return cls._eval(*args, **kwargs)
    
    @classmethod
    def _eval(cls, *values):
        raise NotImplementedError(f"operation not implemented for {cls.__name__}")

    # Reverse operation.
    # For this:
    # c = Operation().call(a, b)
    # This should be true:
    # a = Operation().reverse(b, c)  
    @classmethod
    def _eval_reverse(cls, result, *values):
        raise NotImplementedError(f"reverse operation not implemented for {cls.__name__}")

    @classmethod
    def reverse(cls, result, *values):
        return cls._eval_reverse(result, *values)

    @classmethod
    def has_reverse(cls):
        try:
            cls._eval_reverse(*(2 for _ in range(len(cls._eval_reverse.__code__.co_varnames) - 1)))
            return True
        except NotImplementedError:
            return False        

class EqualOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a == $b"

    @classmethod
    def _eval(cls, a, b):
        return a == b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None

class NotEqualOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a != $b"

    @classmethod
    def _eval(cls, a, b):
        return a != b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return None
        else:
            return b

class GreaterOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a > $b"

    @classmethod
    def _eval(cls, a, b):
        return a > b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None

class GreaterEqualOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a >= $b"

    @classmethod
    def _eval(cls, a, b):
        return a >= b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None

class LessOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a < $b"

    @classmethod
    def _eval(cls, a, b):
        return a < b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None
            
class LessEqualOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a <= $b"

    @classmethod
    def _eval(cls, a, b):
        return a <= b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None


class NumberOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a [number_operator] $b"

    @classmethod
    def _eval(cls, *values):
        return super()._eval(cls)

    @classmethod
    def _eval_reverse(cls, c, b):
        return super()._eval_reverse(c, b)

class AdditionOperator(NumberOperator):
    order = PEMDAS.AS
    print_pattern = "$a + $b"

    @classmethod
    def _eval(cls, a, b):
        return a + b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c - b

class SubtractionOperator(NumberOperator):
    order = PEMDAS.AS
    print_pattern = "$a - $b"

    @classmethod
    def _eval(cls, a, b):
        return a - b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c + b

# All positional operators that have a reverse operation need a
# corresponding backwards operator.
class BackwardsSubtractionOperator(NumberOperator):
    order = PEMDAS.AS
    print_pattern = "$b - $a"

    @classmethod
    def _eval(cls, a, b):
        return b - a
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return b - c

class MultiplicationOperator(NumberOperator):    
    order = PEMDAS.MD
    print_pattern =  "$a * $b"

    @classmethod
    def _eval(cls, a, b):
        return a * b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == 0:
            raise ZeroDivisionError("tried to divide by zero in MultiplicationOperator.reverse")
        return c / b

class DivisionOperator(NumberOperator):
    order = PEMDAS.MD
    print_pattern = "$a / $b"
 
    @classmethod
    def _eval(cls, a, b):
        if b == 0:
            raise ZeroDivisionError("tried to divide by zero in DivisionOperator")
        return a / b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c * b

class BackwardsDivisionOperator(NumberOperator):
    order = PEMDAS.MD
    print_pattern = "$b / $a"

    @classmethod
    def _eval(cls, a, b):
        if a == 0:
            raise ZeroDivisionError("tried to divide by zero in BackwardsDivide")
        return b / a
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return b / c

class PowerOperator(NumberOperator):
    order = PEMDAS.E
    print_pattern = "$a**$b"

    @classmethod
    def _eval(cls, a, b):
        return a ** b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == 0:
            raise ArithmeticError("tried to take zeroth root in PowerOperator.reverse")
        return c ** (1 / b)

class BackwardsPowerOperator(NumberOperator):
    order = PEMDAS.E
    print_pattern = "$b**$a"

    @classmethod
    def _eval(cls, a, b):
        return b ** a
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if c < 0 or b < 0:
            raise ArithmeticError("tried to take logarithm of non positive number in BackwardsPowerOperator.reverse")
        if b == 1:
            raise ArithmeticError("tried divide by zero (log(1)) in BackwardsPowerOperator.reverse")
        return log(c, b)

class RootOperator(NumberOperator):
    order = PEMDAS.E
    print_pattern = "$a**(1/$b)"


    @classmethod
    def _eval(cls, a, b):
        if b == 0:
            raise ArithmeticError("tried to take zeroth root in RootOperator")
        return a ** (1 / b)
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == 0:
            raise ArithmeticError("tried to take zeroth root in RootOperator.reverse")
        return c ** b


class BackwardsRootOperator(NumberOperator):
    order = PEMDAS.E
    print_pattern = "$b**(1/$a)"

    @classmethod
    def _eval(cls, a, b):
        if a == 0:
            raise ArithmeticError("tried to take zeroth root in BackwardsRootOperator")
        return b ** (1 / a)
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b < 0 or c < 0:
            raise ArithmeticError("tried to take logarithm of non positive number in BackwardsRootOperator.reverse")
        if c == 1:
            raise ArithmeticError("tried divide by zero (log(1)) in BackwardsRootOperator.reverse")
        return log(b, c)

class ModuloOperator(NumberOperator):
    order = PEMDAS.MD
    print_pattern = "$a % $b"

    @classmethod
    def _eval(cls, a, b):
        return a % b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        raise NotImplementedError("ModuloOperator does not have a reverse operation")

class AbsoluteOperator(NumberOperator):
    order = PEMDAS.P
    print_pattern = "abs($a)"

    @classmethod
    def _eval(cls, a):
        return abs(a)
    
    @classmethod
    def _eval_reverse(cls, b):
        raise NotImplementedError("AbsOperator does not have a reverse operation")

class NaturalLogarithmOperator(NumberOperator):
    order = PEMDAS.P
    print_pattern = "Ln($a)"

    @classmethod
    def _eval(cls, a):
        return log(a)
    
    @classmethod
    def _eval_reverse(cls, b):
        return exp(b)

class BaseBLogarithmOperator(NumberOperator):
    order = PEMDAS.P
    print_pattern = "LogB($a, $b)"

    @classmethod
    def _eval(cls, a, b):
        return log(a, b)
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return b**c
    
class BackwardsBaseBLogarithmOperator(NumberOperator):
    order = PEMDAS.P
    print_pattern = "RLogB($a, $b)"

    @classmethod
    def _eval(cls, a, b):
        return log(b, a)

    @classmethod
    def _eval_reverse(cls, c, b):
        return exp(log(b) / c)


class BooleanOperator(Operator):
    order = PEMDAS.P
    print_pattern = "$a [boolean_operator] $b"

    @classmethod
    def _eval(cls, *values):
        return super()._eval(cls)

    @classmethod
    def _eval_reverse(cls, result, *values):
        return super()._eval_reverse(result, *values)
        
class AndOperator(BooleanOperator): 
    order = PEMDAS.E
    print_pattern = "$a & $b"

    @classmethod
    def _eval(cls, a, b):
        return a & b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == False:
            return None
        else:
            return c
        
class OrOperator(BooleanOperator): 
    order = PEMDAS.MD
    print_pattern = "$a | $b"

    @classmethod
    def _eval(cls, a, b):
        return a | b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == True:
            return None
        else:
            return c

class XorOperator(BooleanOperator):  
    order = PEMDAS.MD
    print_pattern = "$a ^ $b"
  
    @classmethod
    def _eval(cls, a, b):
        return a ^ b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c ^ b

class NotOperator(BooleanOperator):  
    order = PEMDAS.P
    print_pattern = "~$a"
  
    @classmethod
    def _eval(cls, a):
        if isinstance(a, bool):
            return not a
        return ~a
    
    @classmethod
    def _eval_reverse(cls, b):
        if isinstance(b, bool):
            return not b
        return ~b


import inspect
from typing import List


def get_names(var, frame=None) -> List[str]:
    """
    retrieve all names of a variable in the current frame
    """

    if frame is None:
        frame = inspect.currentframe().f_back
    # Combine global and local variable dicts
    var_dict = frame.f_globals.copy()
    var_dict.update(frame.f_locals)
    return [name for name, value in var_dict.items() if value is var]



import traceback
from ._log import Log
from ..variables import get_instances, get_names
from time import perf_counter
from colorama import Fore, Style


class Timer:
    timers = {}
    auto_log = True
    logging_accuracy = 3

    def __new__(cls, name=None):
        return TimerObject(name)

    @classmethod
    def start(cls, name):
        cls.timers[name] = {}
        cls.timers[name]["start"] = perf_counter()
    
    @classmethod
    def stop(cls, name, auto_log=None):
        if name not in cls.timers:
            raise Exception(f"Timer {name} not started")

        cls.timers[name]["stop"] = perf_counter()
        cls.timers[name]["time"] = cls.timers[name]["stop"] - cls.timers[name]["start"]

        if auto_log is None:
            auto_log = cls.auto_log
        
        if auto_log:
            cls.log(name)
    
    @classmethod
    def log(cls, name):
        if name not in cls.timers:
            raise Exception(f"Timer {name} not started")
        
        if "time" not in cls.timers[name]:
            raise Exception(f"Timer {name} not stopped")

        name_string = f"[Fore: cyan, Style: bright]{name}[Style: reset_all]"

        time_string = f"[Fore: white, Style: bright]{cls.timers[name]['time']:.{cls.logging_accuracy}f}[Style: reset_all]"        

        Log(f"[Style: reset_all]Timer {name_string} took {time_string} seconds")

    @classmethod
    def get(cls, name):
        if name not in cls.timers:
            raise Exception(f"Timer {name} not started")

        if "time" not in cls.timers[name]:
            raise Exception(f"Timer {name} not stopped")

        return cls.timers[name]["time"]

    @classmethod
    def delete(cls, name):
        if name not in cls.timers:
            raise Exception(f"Timer {name} does not exist")
        
        cls.timers.pop(name)

        for object in get_instances(TimerObject):
            if object.name == name:
                object.delete()
    
    @classmethod
    def clear(cls):
        cls.timers = {}


class TimerObject:
    def __init__(self, name):
        if name == None:
            (_,_,_,text)=traceback.extract_stack()[-3]
            name = text[:text.find('=')].strip()
        if name == "Timer(":
            raise Exception("Timer requires a name. Use Timer(name), Timer.start(name) or name = Timer().")
        self.name = name
        self.start()
    
    def start(self):
        Timer.start(self.name)
        return self
    
    def stop(self, auto_log=None):
        Timer.stop(self.name, auto_log)
        return self
    
    def log(self):
        Timer.log(self.name)
        return self
    
    def get(self):
        return Timer.get(self.name)
    
    def delete(self):
        if self.name in Timer.timers:
            Timer.timers.pop(self.name)
        del self